from typing import List, Mapping, Union

# Note: mypy does not yet have support for recursive types, so this will have to do for now
# See discussion here: https://github.com/python/typing/issues/182
_JSONPrimitive = Union[str, float, int, bool, None]
_JSONSerializable1 = Union[_JSONPrimitive, Mapping[str, _JSONPrimitive], List[_JSONPrimitive]]
_JSONSerializable2 = Union[
    _JSONPrimitive, Mapping[str, _JSONSerializable1], List[_JSONSerializable1]
]
_JSONSerializable3 = Union[
    _JSONPrimitive, Mapping[str, _JSONSerializable2], List[_JSONSerializable2]
]
_JSONSerializable4 = Union[
    _JSONPrimitive, Mapping[str, _JSONSerializable3], List[_JSONSerializable3]
]
_JSONSerializable5 = Union[
    _JSONPrimitive, Mapping[str, _JSONSerializable4], List[_JSONSerializable4]
]
JSONSerializable = Union[_JSONPrimitive, Mapping[str, _JSONSerializable5], List[_JSONSerializable5]]



import logging
import time
from functools import wraps

logger = logging.getLogger(__name__)


def retry(
    exceptions=Exception,
    tries=4,
    delay=3,
    backoff=2,
    custom_logger=None,
    verbose=False,
    skip_exceptions=False,
    do_retry_func=None,
):
    """
    Retry calling the decorated function using an exponential backoff.

    :param exceptions: the exception to check. may be a tuple of exceptions to check
    :param tries: number of times to try (not retry) before giving up
    :param delay: initial delay between retries in seconds
    :param backoff: backoff multiplier e.g. value of 2 will double the delay each retry
    :param custom_logger: logger to use. If None, print
    :param verbose: Indicates if the stacktrace should be logged on the first retry
    :param skip_exceptions: Shows whether exceptions will be skipped or not after all tries
    :param do_retry_func: Function that determines if we should retry given the exception
    :return:
    """

    def deco_retry(f):
        @wraps(f)
        def f_retry(*args, **kwargs):
            try_number = 1
            wait_time = delay
            mlogger = custom_logger if custom_logger else logger
            log_stacktrace = verbose

            while try_number < tries:
                try:
                    return f(*args, **kwargs)
                except exceptions as e:
                    if do_retry_func and not do_retry_func(e):
                        mlogger.info('No more retries')
                        if not skip_exceptions:
                            raise e

                    remaining = tries - try_number
                    mlogger.warning(
                        f'{e}, Retrying in {wait_time:.1f} seconds ... (remaining={remaining})',
                        exc_info=log_stacktrace,
                    )

                    time.sleep(wait_time)

                    try_number += 1
                    wait_time = delay * backoff**try_number
                    log_stacktrace = False

            try:
                return f(*args, **kwargs)
            except exceptions as e:
                mlogger.warning('All retries exhausted, terminating.')
                if not skip_exceptions:
                    raise e

        return f_retry  # true decorator

    return deco_retry



from typing import Dict, Optional

from flask import current_app
from mixpanel import Mixpanel

from dh_potluck.types.json_serializable import JSONSerializable

MixpanelProperties = Optional[Dict[str, JSONSerializable]]


class MixpanelService:
    _mp: Mixpanel = None

    @classmethod
    def _get_mixpanel_client(cls):
        if not cls._mp:
            cls._mp = Mixpanel(current_app.config['DH_POTLUCK_MIXPANEL_TOKEN'])
        return cls._mp

    @classmethod
    def send_event(
        cls, user_id: str, event_name: str, properties: MixpanelProperties = None
    ) -> None:
        try:
            cls._get_mixpanel_client().track(user_id, event_name, properties)
        except Exception as e:
            raise MixpanelServiceException(
                message=f'Error occurred while sending mixpanel event: {str(e)}'
            )


class MixpanelServiceException(Exception):
    def __init__(self, message: str = 'Error occurred while sending mixpanel event.') -> None:
        self.message = message
        super().__init__(self.message)



from http import HTTPStatus
from urllib.parse import urljoin

import requests
from flask import current_app


class PlatformConnectionError(Exception):
    pass


class MissingPlatformConnection(PlatformConnectionError):
    pass


class InvalidPlatformConnection(PlatformConnectionError):
    pass


class BadApiResponse(Exception):
    def __init__(self, description, response):
        message = (
            f'{description}\n'
            f'HTTP Status: {response.status_code}\n'
            f'Request URL: {response.url}\n'
            f'Response Body: {response.text}'
        )
        super().__init__(message)


class PlatformConnection:
    @classmethod
    def get(cls, brand_id, platform_type):
        platform_label = platform_type.capitalize()
        res = requests.get(
            urljoin(current_app.config['DH_POTLUCK_AUTH_API_URL'], '/api/platform_connections'),
            headers={
                'Authorization': f'Application {current_app.config["DH_POTLUCK_AUTH_API_TOKEN"]}'
            },
            params={'platform_type': platform_type, 'brand_id': brand_id},
        )
        if not res.ok:
            raise BadApiResponse('Error making API request', res)

        try:
            platform_connection = res.json()[0]

        except IndexError:
            raise MissingPlatformConnection(
                f'No {platform_label} platform connection found for ' f'brand {brand_id}.'
            ) from None

        if platform_connection['status'] != 'connected':
            raise InvalidPlatformConnection(
                f'{platform_label} platform connection for brand '
                f'{brand_id} is marked as '
                f'{platform_connection["status"]}.'
            )

        return platform_connection

    @classmethod
    def update(cls, id, data):
        res = requests.patch(
            urljoin(
                current_app.config['DH_POTLUCK_AUTH_API_URL'], f'/api/platform_connections/{id}'
            ),
            headers={
                'Authorization': f'Application {current_app.config["DH_POTLUCK_AUTH_API_TOKEN"]}'
            },
            json=data,
        )

        if res.status_code != HTTPStatus.OK:
            raise BadApiResponse('Error making API request', res)

        return res.json()



from enum import Enum
from typing import Dict, List, Literal, Optional, TypedDict, Union

import mandrill
from flask import current_app


class MandrillMergeLanguage(Enum):
    """
    Enum for templating languages supported by mailchimp.
    'mailchimp' is the default in Mandrill
    """

    MAILCHIMP = 'mailchimp'
    HANDLEBARS = 'handlebars'


MandrillRecipientType = Union[Literal['to'], Literal['cc'], Literal['bcc']]


class MandrillRecipient(TypedDict):
    email: str
    name: str
    type: MandrillRecipientType


class MandrillEmailClient:
    _DEFAULT_EMAIL_PROPS = {
        'from_name': 'Dash Hudson',
        'from_email': 'noreply@dashhudson.com',
        'preserve_recipients': True,
        'track_clicks': True,
        'track_opens': True,
        'inline_css': True,
    }

    _mandrill_client: mandrill.Mandrill

    def __init__(self):
        self._mandrill_client = mandrill.Mandrill(current_app.config['DH_POTLUCK_MANDRILL_API_KEY'])

    def send_email_template(
        self,
        recipients: List[MandrillRecipient],
        template_name: str,
        template_vars: List[Dict[str, str]],
        subject: Optional[str] = None,
        merge_language: Optional[MandrillMergeLanguage] = None,
    ) -> None:
        message = {
            'preserve_recipients': self._DEFAULT_EMAIL_PROPS['preserve_recipients'],
            'track_clicks': self._DEFAULT_EMAIL_PROPS['track_clicks'],
            'track_opens': self._DEFAULT_EMAIL_PROPS['track_opens'],
            'from_name': self._DEFAULT_EMAIL_PROPS['from_name'],
            'from_email': self._DEFAULT_EMAIL_PROPS['from_email'],
            'to': recipients,
            'global_merge_vars': template_vars,
        }
        if subject:
            message['subject'] = subject

        if merge_language:
            message['merge_language'] = merge_language.value

        self._mandrill_client.messages.send_template(
            template_name=template_name, message=message, template_content=[]
        )

    def send_email_html(
        self,
        recipients: List[MandrillRecipient],
        subject: Optional[str] = None,
        html: Optional[str] = None,
    ) -> None:
        message = {
            'preserve_recipients': self._DEFAULT_EMAIL_PROPS['preserve_recipients'],
            'track_clicks': self._DEFAULT_EMAIL_PROPS['track_clicks'],
            'track_opens': self._DEFAULT_EMAIL_PROPS['track_opens'],
            'from_name': self._DEFAULT_EMAIL_PROPS['from_name'],
            'from_email': self._DEFAULT_EMAIL_PROPS['from_email'],
            'inline_css': self._DEFAULT_EMAIL_PROPS['inline_css'],
            'to': recipients,
        }
        if html:
            message['html'] = html

        if subject:
            message['subject'] = subject

        self._mandrill_client.messages.send(message=message)



import json
import logging
import re
from urllib.parse import urlencode

import json_log_formatter
from boltons.iterutils import remap
from ddtrace import tracer
from flask import request

SENSITIVE_KEYS = re.compile(r'password|token|secret|key', flags=re.I)
MAX_BODY_SIZE = 50000


def set_log_levels(config_object):
    general_log_level = getattr(config_object, 'DH_POTLUCK_LOG_LEVEL', 'INFO')
    if general_log_level:
        try:
            numeric_level = _get_numeric_log_level(general_log_level.upper())
        except ValueError as e:
            raise ValueError(f'General logging configuration error: {e}')
        logging.basicConfig(level=numeric_level)

    module_log_levels = getattr(config_object, 'DH_POTLUCK_MODULE_LOG_LEVELS', None)
    if module_log_levels:
        module_configs = dict(
            module_config.split(':')
            for module_config in module_log_levels.split(',')
            if module_config and len(module_config.split(':')) == 2
        )
        for module, log_level in module_configs.items():
            module = module.strip()
            log_level = log_level.strip().upper()
            try:
                numeric_level = _get_numeric_log_level(log_level)
            except ValueError as e:
                raise ValueError(f'Module "{module}" logging configuration error: {e}')
            logging.getLogger(module).setLevel(numeric_level)


def _get_numeric_log_level(log_level: str):
    numeric_level = getattr(logging, log_level, None)
    if not isinstance(numeric_level, int):
        raise ValueError(f'Invalid log level: {log_level}')
    return numeric_level


def add_request_params_to_trace():
    """
    Add additional details about the current request (query string, body, request size) to
    the current Datadog trace. Useful for error handlers.
    """
    span = tracer.current_root_span()
    if not span:
        return

    # Log query string (if present) for all request methods
    query_params = request.args
    if query_params:
        clean = remap(query_params.copy(), visit=scrub_keys)
        span.set_tag('http.query_string', urlencode(clean))

    # Skip body logging if not POST, PATCH or PUT
    if request.method not in ['POST', 'PATCH', 'PUT']:
        return

    # Skip body logging if it's empty
    if not request.content_length:
        return

    span.set_tag('http.content_length', str(request.content_length))

    if request.content_length > MAX_BODY_SIZE:
        span.set_tag('http.body', 'Body too large, content could not be logged.')
        return

    # Try to parse body as JSON, and scrub sensitive values
    body = request.get_json(silent=True)
    if body:
        clean = remap(body, visit=scrub_keys)
        span.set_tag('http.body', json.dumps(clean))
    else:
        # If we can't parse as JSON, log the raw body
        body = request.get_data(as_text=True)
        span.set_tag('http.body', body)


def scrub_keys(path, key, value):
    if isinstance(key, str) and SENSITIVE_KEYS.search(key):
        return key, '-' * 5
    return key, value


class DHJSONFormatter(json_log_formatter.JSONFormatter):
    def json_record(self, message, extra, record):
        extra['funcName'] = record.funcName
        extra['level'] = record.levelname
        extra['lineno'] = record.lineno
        extra['pathname'] = record.pathname
        extra['module'] = record.module
        return super(DHJSONFormatter, self).json_record(message, extra, record)


def structure_logger(logger):
    # Assign JSON formatter to handler
    json_handler = logging.StreamHandler()
    json_handler.setFormatter(DHJSONFormatter())
    logger.handlers = [json_handler]
    logger.propagate = False


def patch_celery_get_logger():
    # This has the opportunity of being a patch closer to the source (kombu)
    # Imports safely hidden behind structured_logging flag for now
    try:
        from celery.utils import log

        def _patch_func(name):
            l = log._get_logger(name)  # noqa: E741
            if logging.root not in (l, l.parent) and l is not log.base_logger:
                l = log._using_logger_parent(log.base_logger, l)  # noqa: E741
            structure_logger(l)
            return l

        log.get_logger = _patch_func
    except ModuleNotFoundError:
        return



from typing import Dict, List, Optional

from dh_potluck.email.mandrill_email_client import (
    MandrillEmailClient,
    MandrillMergeLanguage,
    MandrillRecipient,
)


class EmailService(object):
    _email_client: Optional[MandrillEmailClient] = None

    @classmethod
    def _get_email_client(cls):
        if not cls._email_client:
            cls._email_client = MandrillEmailClient()
        return cls._email_client

    @classmethod
    def send_email_template(
        cls,
        recipients: List[MandrillRecipient],
        template_name: str,
        template_vars: List[Dict[str, str]],
        subject: Optional[str] = None,
        merge_language: Optional[MandrillMergeLanguage] = None,
    ) -> None:
        try:
            cls._get_email_client().send_email_template(
                recipients,
                template_name,
                template_vars,
                subject=subject,
                merge_language=merge_language,
            )
        except Exception as e:
            raise EmailServiceException(
                message=f'Error occurred while sending email template: {str(e)}'
            )

    @classmethod
    def send_email_html(
        cls,
        recipients: List[MandrillRecipient],
        subject: Optional[str] = None,
        html: Optional[str] = None,
    ) -> None:
        try:
            cls._get_email_client().send_email_html(recipients, subject=subject, html=html)
        except Exception as e:
            raise EmailServiceException(
                message=f'Error occurred while sending html email: {str(e)}'
            )


class EmailServiceException(Exception):
    def __init__(self, message: str = 'Error occurred while sending email.') -> None:
        self.message = message
        super().__init__(self.message)



from datetime import datetime
from typing import TypedDict

from confluent_kafka.cimpl import Message as KafkaMessage


class MessageValue(TypedDict):
    message_type: str
    message_time: datetime
    payload: str


class Message(object):
    _topic: str
    _value: MessageValue
    _raw_message: KafkaMessage

    def __init__(self, topic: str, value: MessageValue, raw_message: KafkaMessage):
        """
        :param topic: The topic the message came from
        :param value: The message value
        :param raw_message: The raw kafka message object
        """
        self._topic = topic
        self._value = value
        self._raw_message = raw_message

    def topic(self) -> str:
        return self._topic

    def value(self) -> MessageValue:
        return self._value

    def raw_message(self) -> KafkaMessage:
        return self._raw_message



import hashlib
import logging
import re
from functools import wraps
from http import HTTPStatus
from pickle import dumps, loads

import requests
from flask import current_app, g, jsonify, request
from flask_redis import FlaskRedis
from redis.exceptions import ResponseError

logger = logging.getLogger(__name__)


token_missing = {
    'description': 'Authentication token missing or incorrectly formatted.',
    'status': HTTPStatus.UNAUTHORIZED,
}
token_invalid = {
    'description': 'Authentication token invalid or expired.',
    'status': HTTPStatus.UNAUTHORIZED,
}
user_disabled = {
    'description': 'User account disabled.',
    'status': HTTPStatus.FORBIDDEN,
}
permission_denied = {
    'description': 'You do not have access to this resource.',
    'status': HTTPStatus.FORBIDDEN,
}
auth_error = {
    'description': 'An error occurred trying to authenticate.',
    'status': HTTPStatus.INTERNAL_SERVER_ERROR,
}


def error_response(error):
    return jsonify({'description': error['description']}), error['status']


class UnauthenticatedUser:
    role = None
    is_active = True


class AuthenticatedUser:
    id = None
    email = None
    first_name = None
    last_name = None
    default_time_zone = None
    time_zone_name = None
    password_reset_expires = None
    password_reset_url = None
    has_device = None
    status = None
    brandpanel_id = None
    is_superadmin = None
    is_pending = None
    job_title = None
    avatar_url = None
    organization = None
    brands = None
    accessible_brands = None
    permissions = None
    updated_at = None
    created_at = None

    def __init__(self, *initial_data, **kwargs):
        for dictionary in initial_data:
            for key in dictionary:
                setattr(self, key, dictionary[key])
        for key in kwargs:
            setattr(self, key, kwargs[key])

    @property
    def role(self):
        if self.is_superadmin:
            return 'superadmin'
        return 'user'

    @property
    def is_active(self):
        return self.status == 0


class ApplicationUser:
    role = 'app'
    is_active = True


def get_user(app_token, validate_token_func):
    auth_header = request.headers.get('Authorization')
    if auth_header is None:
        g.user = UnauthenticatedUser()
        return

    match = re.match(r'^(Application|Token|Bearer):? (\S+)$', auth_header)
    if not match:
        return error_response(token_missing)
    method = match.group(1)
    token = match.group(2)

    # Application token
    # Note: 'Token' is deprecated - all applications should be updated to use 'Application'
    if method == 'Application' or method == 'Token':
        if token == app_token:
            g.user = ApplicationUser()
        else:
            return error_response(token_invalid)

    # Bearer token
    elif method == 'Bearer':
        user = validate_token_func(match.group(2))
        if not user:
            return error_response(token_invalid)
        g.user = user

    if not g.user.is_active:
        return error_response(user_disabled)


roles = {
    'user': 0,
    'brand_admin': 1,
    'superadmin': 2,
    'app': 3,
}


def role_required(role):
    """
    Currently, the supported roles are:

    1. user
    2. brand_admin
    3. superadmin
    4. app

    Roles are ordered from least to most privilege. Each role receives the permissions of the roles
    before it. Example:

    If role='user', users with 'user', 'superadmin', or 'app' roles will be granted access.
    If role='superadmin', only users with 'superadmin' or 'app' roles will be granted access.
    """

    def decorator(func):
        # finds the role within a brand of the current user
        def get_brand_role():
            # extract the brand_id from the path
            brand_id = request.view_args.get('brand_id')
            if not brand_id:
                return None

            # use the brand_id to find the specific brand_role from the list of brand_roles
            brs = g.user.permissions.get('brand_roles', [])
            br = next((br.get('role') for br in brs if br.get('brand_id') == brand_id), None)

            # br.value will exist if g.user is assigned a user model (called from auth)
            # otherwise, br alone is the value we want
            return getattr(br, 'value', br)

        @wraps(func)
        def wrapper(*args, **kwargs):
            if isinstance(g.user, UnauthenticatedUser):
                return error_response(token_missing)

            if roles[g.user.role] >= roles[role]:
                return func(*args, **kwargs)

            # a check for the brand_admin role is needed as it isn't stored in the user instance
            if role == 'brand_admin' and g.user.brands and get_brand_role() == 'admin':
                return func(*args, **kwargs)

            return error_response(permission_denied)

        return wrapper

    return decorator


class AuthCache(object):
    cache = None

    @classmethod
    def get_cache(cls):
        if not current_app.config.get('DH_POTLUCK_AUTH_REDIS_URL'):
            logger.debug('Identity cache url (DH_POTLUCK_AUTH_REDIS_URL) not configured')
            return None
        if not cls.cache:
            cls.cache = FlaskRedis(
                app=current_app, config_prefix='DH_POTLUCK_AUTH_REDIS', socket_timeout=0.4
            )
        return cls.cache


def _get_from_cache(token):
    logger.debug('Get identity from cache')
    cache = AuthCache.get_cache()
    if not cache:
        return None
    try:
        cached_identity = cache.get(get_auth_cache_key(token))
        if cached_identity:
            logger.debug('Identity found!')
            return loads(cached_identity)
    except TimeoutError as err:
        logger.error(err)
        return None


def _set_to_cache(token, identity):
    logger.debug('Set identity to cache')
    ex = current_app.config.get('DH_POTLUCK_IDENTITY_EXPIRY_IN_SECONDS') or (5 * 60)
    cache = AuthCache.get_cache()
    if not cache:
        return
    try:
        cache.set(get_auth_cache_key(token), dumps(identity), ex=ex)
    except (ResponseError, TimeoutError) as err:
        logger.error(err)


def get_auth_cache_key(token):
    hash_object = hashlib.sha1(token.encode('utf-8'))
    token_hash = hash_object.hexdigest()
    return f'auth:token#{token_hash}'


def validate_token_using_api(token):
    if current_app.config.get('DH_POTLUCK_ENABLE_AUTH_CACHE'):
        identity = _get_from_cache(token)
        if identity:
            return AuthenticatedUser(identity)

    logger.debug('Get identity from auth service')
    auth_api_url = current_app.config['DH_POTLUCK_AUTH_API_URL']
    res = requests.get(
        auth_api_url + 'self',
        headers={'Authorization': f'Bearer {token}', 'content-type': 'application/json'},
    )
    if res.status_code == HTTPStatus.OK:
        identity = res.json()
        if current_app.config.get('DH_POTLUCK_ENABLE_AUTH_CACHE'):
            # TODO: Remove this once auth is upgraded to use the new cache key
            _set_to_cache(token, identity)
        return AuthenticatedUser(identity)

    return None



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['dh_potluck',
 'dh_potluck.celery',
 'dh_potluck.email',
 'dh_potluck.messaging',
 'dh_potluck.types',
 'dh_potluck.utils']

package_data = \
{'': ['*'], 'dh_potluck': ['templates/*']}

install_requires = \
['Werkzeug>=0.15.0',
 'boltons>=20.2.1,<21.0.0',
 'boto3>=1.17.102,<2.0.0',
 'confluent-kafka>=1.6.0,<2.0.0',
 'ddtrace==1.1.1',
 'flask-limiter>=2.2.0,<3.0.0',
 'flask-redis>=0.4.0,<0.5.0',
 'flask-smorest>=0.38.1,<0.39.0',
 'flask>=2.1.2,<3.0.0',
 'flask_sqlalchemy>=2.4.4,<3.0.0',
 'json-log-formatter>=0.3.0,<0.4.0',
 'mandrill>=1.0.60,<2.0.0',
 'marshmallow==3.17.0',
 'mixpanel>=4.9.0,<5.0.0',
 'requests>=2.22,<3.0',
 'sqlalchemy>=1.3,<2.0']

setup_kwargs = {
    'name': 'dh-potluck',
    'version': '1.7.0',
    'description': '',
    'long_description': None,
    'author': None,
    'author_email': None,
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)



from typing import Optional


class SemaphoreLocked(Exception):
    pass


try:
    from redis import Redis

    class Semaphore(object):
        """
        Class allowing using Redis backend for task synchronization

        Synopsis:

        redis = Redis(host=..., port=..., db=...)

        # mutual exclusivity (mutex) to perform operation on a foo resource
        @celery.task
        def a_task(foo, bar):
            with Semaphore(redis, 'mutex'+foo, 60):
                # now I have exclusive access to for 60 seconds to 'foo' resource
                do_the_job()
            except MutexTaken:
                # some other task is already using the resource so I skip
                pass

        # optionally restrict to only N concurrent operations on a specific task, regardless of the
        # task capacity (warning, will be blocking tasks waiting on execution)

        def optional_limited_capacity():
            if not Config.MAX_CONCURRENT_RUNNING_TASKS:
                # always allowing entering the resource
                return Semaphore.dummy()
            return Semaphore(redis, 'throttled_resource', 3600, Config.MAX_CONCURRENT_RUNNING_TASKS)

        @celery.task
        def throttled_task(foo, bar):
            while True:
                with optional_limited_capacity():
                    # now I am one of N to do the work
                    do_the_job();
                    break;
                except MutexTaken:
                    time.sleep(1) # not to hog server CPU

        """

        def __init__(
            self,
            redis_client: Optional[Redis],
            key_name: Optional[str],
            ttl: int,
            capacity: int = 1,
        ):
            if redis_client is not None and key_name is not None:
                current_value = int(redis_client.get(key_name) or 0)
                if current_value >= capacity:
                    raise SemaphoreLocked()
                current_value += 1
                redis_client.setex(key_name, time=ttl, value=current_value)
            self._key_name = key_name
            self._ttl = ttl
            self._client = redis_client

        def __enter__(self):
            return self

        def __exit__(self, exc_type, exc_val, exc_tb):
            if self._key_name is not None:
                current_value = int(self._client.get(self._key_name) or 0)
                current_value -= 1
                if current_value <= 0:
                    self._client.delete(self._key_name)
                else:
                    self._client.setex(self._key_name, time=self._ttl, value=current_value)

        @classmethod
        def dummy(cls) -> 'Semaphore':
            return Semaphore(None, None, 0)

except ModuleNotFoundError:
    pass
